"""Contains the MagicTodoMixin class definition."""

from __future__ import annotations

import abc
import datetime as dt
from functools import total_ordering
from typing import Any, Dict, Generic, List, Tuple, Type

from eris import ErisError, Err, Ok, Result

from . import types as mtypes
from ._todo import Todo, TodoMixin
from .spells import POST_BUILTIN_SPELLS, PRE_BUILTIN_SPELLS


@total_ordering
class MagicTodoMixin(TodoMixin, Generic[mtypes.MagicTodo_T], abc.ABC):
    """Mixin class that implements the Todo protocol."""

    pre_spells: List[mtypes.TodoSpell] = PRE_BUILTIN_SPELLS
    post_spells: List[mtypes.TodoSpell] = POST_BUILTIN_SPELLS

    def __init__(self: mtypes.MagicTodo_T, todo: Todo):
        self._todo = todo
        self.todo = self.cast_spells(todo).unwrap()

    @classmethod
    def from_line(
        cls: Type[mtypes.MagicTodo_T], line: str
    ) -> Result[mtypes.MagicTodo_T, ErisError]:
        """Converts a string into a MagicTodo object."""
        todo_result = Todo.from_line(line)

        err: Err[Any, ErisError]
        if isinstance(todo_result, Err):
            err = Err(
                "Failed to construct basic Todo object inside of MagicTodo."
            )
            return err.chain(todo_result)

        todo = todo_result.ok()
        if error := cls.cast_spells(todo).err():
            err = Err(f"Failed spell validation for this todo: todo={todo!r}")
            return err.chain(error)

        return Ok(cls(todo))

    def to_line(self) -> str:
        """Converts this MagicTodo back to a string."""
        return self.todo.to_line()

    def to_dict(self: mtypes.MagicTodo_T) -> dict[str, Any]:
        """Converts this MagicTodo into a dictionary."""
        result: Dict[str, Any] = {}
        result["_todo"] = self.todo.to_dict()
        result["todo"] = self.todo.to_dict()
        result["pre_spells"] = [spell.__name__ for spell in self.pre_spells]
        result["spells"] = [spell.__name__ for spell in self.spells]
        result["post_spells"] = [spell.__name__ for spell in self.post_spells]
        return result

    @classmethod
    def cast_spells(
        cls: Type[mtypes.MagicTodo_T], todo: Todo
    ) -> Result[Todo, ErisError]:
        """Casts all spells associated with this MagicTodo on `todo`."""
        new_todo = todo.new()
        for spell_list in [cls.pre_spells, cls.spells, cls.post_spells]:
            for spell in spell_list:
                new_todo_result = spell(new_todo)
                if isinstance(new_todo_result, Err):
                    err: Err[Any, ErisError] = Err(
                        f"The {spell.__name__!r} spell failed while processing"
                        " this todo."
                    )
                    return err.chain(new_todo_result)

                new_todo = new_todo_result.ok()

        return Ok(new_todo)

    @property
    def contexts(self) -> Tuple[str, ...]:  # noqa: D102
        return self.todo.contexts

    @property
    def create_date(self) -> dt.date | None:  # noqa: D102
        return self.todo.create_date

    @property
    def desc(self) -> str:  # noqa: D102
        return self.todo.desc

    @property
    def done_date(self) -> dt.date | None:  # noqa: D102
        return self.todo.done_date

    @property
    def marked_done(self) -> bool:  # noqa: D102
        return self.todo.marked_done

    @property
    def metadata(self) -> mtypes.Metadata:  # noqa: D102
        return self.todo.metadata

    @property
    def priority(self) -> mtypes.Priority:  # noqa: D102
        return self.todo.priority

    @property
    def projects(self) -> Tuple[str, ...]:  # noqa: D102
        return self.todo.projects


class MagicTodo(MagicTodoMixin):
    """The default MagicTodo class."""

    spells: List[mtypes.TodoSpell] = []



"""Test that we can create custom MagicTodo classes."""

from __future__ import annotations

from typing import Type

from pytest import mark

from magodo import MagicTodo, Todo
from magodo._shared import to_date
from magodo.types import AbstractMagicTodo


params = mark.parametrize


@params(
    "line,magic_todo_type,todo,etodo",
    [
        (
            "x:1030 2022-01-12 Some done todo.",
            MagicTodo,
            Todo(
                desc="x:1030 2022-01-12 Some done todo.",
                metadata={"x": "1030"},
            ),
            Todo(
                desc="Some done todo. | dtime:1030",
                marked_done=True,
                create_date=to_date("2022-01-12"),
                metadata={"dtime": "1030"},
            ),
        ),
        (
            "x:1030 2022-01-12 Some done todo with a @ctx.",
            MagicTodo,
            Todo(
                desc="x:1030 2022-01-12 Some done todo with a @ctx.",
                contexts=("ctx",),
                metadata={"x": "1030"},
            ),
            Todo(
                desc="Some done todo with a ctx. | @ctx dtime:1030",
                contexts=("ctx",),
                marked_done=True,
                create_date=to_date("2022-01-12"),
                metadata={"dtime": "1030"},
            ),
        ),
    ],
)
def test_magic_todo(
    line: str,
    magic_todo_type: Type[AbstractMagicTodo],
    todo: Todo,
    etodo: Todo,
) -> None:
    """Test the MagicTodo.from_line() function."""
    actual = magic_todo_type.from_line(line).unwrap()
    expected = magic_todo_type(todo)
    assert expected.todo == etodo
    assert actual.to_dict() == expected.to_dict()



"""Contains all of magodo's default MagicTodo spell functions."""

from __future__ import annotations

from typing import Callable, Iterable, List, MutableSequence

from eris import ErisError, Err, Ok, Result

from ._shared import (
    CONTEXT_PREFIX,
    DEFAULT_PRIORITY,
    PROJECT_PREFIX,
    PUNCTUATION,
    TODO_PREFIXES,
    is_any_prefix_tag,
    is_any_tag,
    is_metadata_tag,
)
from ._todo import Todo
from .types import TodoSpell


def register_spell_factory(
    mut_spell_registry: MutableSequence[TodoSpell],
) -> Callable[[TodoSpell], TodoSpell]:
    """Factory for decorators used to register spell functions."""

    def register_spell(spell: TodoSpell) -> TodoSpell:
        mut_spell_registry.append(spell)
        return spell

    return register_spell


PRE_BUILTIN_SPELLS: List[TodoSpell] = []
pre_spell = register_spell_factory(PRE_BUILTIN_SPELLS)

POST_BUILTIN_SPELLS: List[TodoSpell] = []
post_spell = register_spell_factory(POST_BUILTIN_SPELLS)


@pre_spell
def handle_prefix(todo: Todo) -> Result[Todo, ErisError]:
    """Handles / validatees a magic todo's prefix.."""
    if todo.priority == DEFAULT_PRIORITY and not todo.desc.startswith(
        TODO_PREFIXES
    ):
        return Err(
            "Magic todos must satisfy one of; (1) Have a non-default"
            " (i.e. not 'O') priority set (2) Have been marked complete"
            " with an 'x' prefix (3) Have been marked open with an 'o'"
            f" prefix. todo={todo!r}"
        )

    if todo.desc.startswith("o "):
        return Ok(todo.new(desc=todo.desc[2:]))

    return Ok(todo)


@pre_spell
def x_tag(todo: Todo) -> Result[Todo, ErisError]:
    """Handles tags of the form x:1234 where 1234 is the current time."""
    if todo.metadata is None or "x" not in todo.metadata:
        return Ok(todo)

    dtime = todo.metadata["x"]
    desc = " ".join(todo.desc.split(" ")[1:]) + f" dtime:{dtime}"

    new_todo = todo.new(desc=desc, marked_done=True)
    line = new_todo.to_line()
    return Todo.from_line(line)


@post_spell
def group_tags(todo: Todo) -> Result[Todo, ErisError]:
    """Groups all @ctxs, +projs, and meta:data at the end of the line."""
    if not (todo.contexts or todo.projects or todo.metadata):
        return Ok(todo)

    all_words = [w for w in todo.desc.split(" ") if w != "|"]
    new_words = []
    regular_words_found = False
    for i, word in enumerate(all_words[:]):
        all_prev_words_are_tags = _all_words_are_tags(all_words[:i])
        all_next_words_are_tags = _all_words_are_tags(all_words[i + 1 :])
        is_edge_tag = all_next_words_are_tags or all_prev_words_are_tags

        if is_metadata_tag(word) and word[-1] in PUNCTUATION:
            return Ok(todo)

        if is_any_prefix_tag(word) and (
            word[-1] in PUNCTUATION or not all_next_words_are_tags
        ):
            if regular_words_found:
                new_words.append(word[1:])
            continue

        if is_any_prefix_tag(word) and is_edge_tag:
            continue

        if is_metadata_tag(word):
            continue

        regular_words_found = True
        new_words.append(word)

    if not regular_words_found:
        return Ok(todo)

    desc = " ".join(new_words).strip()
    desc += " |"

    if todo.contexts:
        desc += " " + " ".join(
            CONTEXT_PREFIX + ctx for ctx in sorted(todo.contexts)
        )

    if todo.projects:
        desc += " " + " ".join(
            PROJECT_PREFIX + ctx for ctx in sorted(todo.projects)
        )

    if todo.metadata:
        desc += " " + " ".join(
            f"{k}:{v}" for (k, v) in sorted(todo.metadata.items())
        )

    return Ok(todo.new(desc=desc))


def _all_words_are_tags(words: Iterable[str]) -> bool:
    """Returns True if all `words` are special words."""
    return all(is_any_tag(w) for w in words)



"""Contains the TodoGroup class definition."""

from __future__ import annotations

import datetime as dt
from logging import Logger
from pathlib import Path
from typing import Generic, Iterable, Iterator, List, Type

from eris import Err
from metaman import cname
from typist import PathLike

from . import types as mtypes


logger = Logger(__name__)


class TodoGroup(Generic[mtypes.Todo_T]):
    """Manages a group of Todo objects."""

    def __init__(self, todos: Iterable[mtypes.Todo_T]) -> None:
        self._todos = sorted(todos)

    def __repr__(self) -> str:  # noqa: D105
        return f"{cname(self)}({self._todos})"

    def __iter__(self) -> Iterator[mtypes.Todo_T]:
        """Yields the Todo objects that belong to this group."""
        yield from self._todos

    def __len__(self) -> int:  # noqa: D105
        return len(self._todos)

    @classmethod
    def from_path(
        cls, todo_type: Type[mtypes.Todo_T], path_like: PathLike
    ) -> TodoGroup:
        """Reads all todo lines from a given file or directory (recursively).

        Pre-conditions:
            * `path_like` exists and is either a file or directory.
        """
        path = Path(path_like)

        assert path.exists(), f"The provided path does not exist: {path}"

        todos: List[mtypes.Todo_T] = []
        if path.is_file():
            logger.debug(
                "Attempting to load todos from text file: file=%r", str(path)
            )

            for line in path.read_text().split("\n"):
                line = line.strip()
                todo_result = todo_type.from_line(line)
                if isinstance(todo_result, Err):
                    continue

                todo = todo_result.ok()
                todos.append(todo)
                logger.debug("New todo loaded: todo=%r", todo)
        else:
            assert path.is_dir(), (
                "The provided path exists but is neither a file nor a"
                f" directory: {path}"
            )

            logger.debug("Loading from directory: dir=%r", str(path))
            for other_path in path.glob("*"):
                if other_path.is_file() and other_path.suffix != ".txt":
                    continue

                other_todo_group = TodoGroup.from_path(todo_type, other_path)
                todos.extend(other_todo_group)

        return cls(todos)

    def filter_by(
        self,
        *,
        contexts: Iterable[str] = None,
        create_date: dt.date = None,
        desc: str = None,
        done_date: dt.date = None,
        marked_done: bool = None,
        metadata: mtypes.Metadata = None,
        priority: mtypes.Priority = None,
        projects: Iterable[str] = None,
    ) -> TodoGroup:
        """Filter this group using one or more Todo properties."""
        todos = []

        for todo in self._todos:
            if contexts is not None and not all(
                ctx in todo.contexts for ctx in contexts
            ):
                continue

            if create_date is not None and todo.create_date != create_date:
                continue

            if desc is not None and desc.lower() not in todo.desc.lower():
                continue

            if done_date is not None and todo.done_date != done_date:
                continue

            if marked_done is not None and todo.marked_done != marked_done:
                continue

            if metadata is not None and not all(
                v == todo.metadata.get(k, None) for (k, v) in metadata.items()
            ):
                continue

            if priority is not None and todo.priority != priority:
                continue

            if projects is not None and not all(
                proj in todo.projects for proj in projects
            ):
                continue

            todos.append(todo)

        return TodoGroup(todos)
